{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "tags": [
     "hide-cell"
    ]
   },
   "outputs": [],
   "source": [
    "\"\"\"Beryllium atom\"\"\"\n",
    "\n",
    "__author__    = \"Magnar Bjørgve\"\n",
    "__credit__    = [\"Magnar Bjørgve\", \"Stig Rune Jensen\"]\n",
    "\n",
    "__date__      = \"2021-02-16\""
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Beryllium atom"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We solve the Hartree-Fock equations for a closed-shell many-electron system\n",
    "\\begin{equation*}\n",
    "\\left[\\hat{T} + \\hat{V}\\right]\\Phi = \\Phi F\n",
    "\\end{equation*}\n",
    "\n",
    "In the equation above $\\Phi$ is a row vector of doubly occupied orbitals,\n",
    "$F = \\langle\\Phi|\\hat{T}+\\hat{V}|\\Phi\\rangle$ is the Fock matrix, and\n",
    "$\\hat{V} = \\hat{V}_{nuc} + 2\\hat{J} - \\hat{K}$ is the potential operator.\n",
    "We here define the nuclear attraction potential\n",
    "\\begin{align*}\n",
    "\\hat{V}_{nuc}(r) = \\sum_I \\frac{-Z_I}{|R_I - r|}\n",
    "\\end{align*}\n",
    "with $Z_I$ and $R_I$ being the nuclear charges and positions,\n",
    "and the Coulomb and Exchange operators\n",
    "\\begin{align*}\n",
    "\\hat{J} \\varphi_i &= \\sum_j \\hat{P}[\\varphi_j \\varphi_j^\\dagger] \\varphi_i \\\\\n",
    "\\hat{K} \\varphi_i &= \\sum_j \\hat{P}[\\varphi_i \\varphi_j^\\dagger] \\varphi_j\n",
    "\\end{align*}\n",
    "where $\\hat{P}$ is the convolution operator with the Poisson kernel $P=1/r$.\n",
    "We iterate the Hartree-Fock equations in their integral form\n",
    "\\begin{equation*}\n",
    "\\tilde{\\Phi}^{n+1} = -2 \\hat{G}^{\\vec{\\mu}^n}\\left[\\hat{V} \\Phi^n + \\Phi^n\\left(\\Lambda^n - F^n\\right)\\right]\n",
    "\\end{equation*}\n",
    "where $\\hat{G}^{\\vec{\\mu}}$ is a vector of convolution operators with the\n",
    "bound-state Helmholtz kernel $G^{\\mu} = e^{-\\mu r}/r$ and $\\mu_i^n = \\sqrt{-2\\lambda_i^n}$\n",
    "is a positive real number. $\\Lambda_{ij}^n = \\lambda_i^n\\delta_{ij}$ is a diagonal matrix containing\n",
    "the same parameters as used in the construction of the Helmholz vector. In practice we choose\n",
    "$\\lambda_i^n = F_{ii}^n$ such that diagonal elements vanish in $\\left(\\Lambda^n - F^n\\right)$."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "tags": [
     "remove_cell"
    ]
   },
   "source": [
    "<div style=\"background: #efffed;\n",
    "            border: 1px solid grey\n",
    "            margin: 8px 0 8px 0;\n",
    "            text-align: center;\n",
    "            padding: 8px;\">\n",
    "    <i class=\"fa-play fa\"\n",
    "        style=\"font-size: 40px;\n",
    "              line-height: 40px;\n",
    "              margin: 8px;\n",
    "              color: #444;\" >\n",
    "    </i>\n",
    "    <div>\n",
    "    To run the selected code cell, hit <pre style=\"background: #efffed\">Shift + Enter</pre>\n",
    "    </div>\n",
    "<div>"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "from vampyr import vampyr3d as vp\n",
    "import numpy as np\n",
    "import numpy.linalg as LA\n",
    "import matplotlib.pyplot as plt"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "class HelmholtzOperator():\n",
    "    \"\"\"\n",
    "    Vectorized Helmholtz operator\n",
    "\n",
    "    Parameters\n",
    "    ----------\n",
    "    mra : The multiresolution analysis we work on\n",
    "    lamb : vector of lambda parameters, mu_i = sqrt(-2*lambda_i)\n",
    "    prec : Precision requirement\n",
    "\n",
    "    Attributes\n",
    "    ----------\n",
    "    operators : list containing HelmholtzOperators for each orbital\n",
    "\n",
    "    \"\"\"\n",
    "    def __init__(self, mra, lamb, prec):\n",
    "        self.mra = mra\n",
    "        self.lamb = lamb\n",
    "        self.prec = prec\n",
    "        self.operators = []\n",
    "        self.setup()\n",
    "\n",
    "    def setup(self):\n",
    "        mu = [np.sqrt(-2.0*l) if l < 0 else 1.0 for l in self.lamb]\n",
    "        for m in mu:\n",
    "            self.operators.append(vp.HelmholtzOperator(mra=self.mra, exp=m, prec=self.prec))\n",
    "\n",
    "    def __call__(self, Psi):\n",
    "        \"\"\"Operate the Helmholtz operator onto an orbital vector\"\"\"\n",
    "        return np.array([self.operators[i](Psi[i]) for i in range(len(Psi))])\n",
    "\n",
    "class ExchangeOperator():\n",
    "    \"\"\"\n",
    "    Vectorized Exchange operator\n",
    "\n",
    "    K(phi) = \\sum_i P[phi * psi_i] * psi_i\n",
    "\n",
    "    Parameters\n",
    "    ----------\n",
    "    mra : The multiresolution analysis we work on\n",
    "    Psi : Orbital vector defining the operator\n",
    "    prec : Precision requirement\n",
    "\n",
    "    Attributes\n",
    "    ----------\n",
    "    poisson : Poisson Operator\n",
    "\n",
    "    \"\"\"\n",
    "    def __init__(self, mra, Psi, prec):\n",
    "        self.mra = mra\n",
    "        self.Psi = Psi\n",
    "        self.prec = prec\n",
    "        self.poisson = vp.PoissonOperator(mra=mra, prec=self.prec)\n",
    "\n",
    "    def __call__(self, Phi):\n",
    "        \"\"\"Apply the exchange operator onto an orbital vector Phi\"\"\"\n",
    "\n",
    "        Phi_out = []\n",
    "        for j in range(len(Phi)):\n",
    "            V_j0 = self.poisson(Phi[j] * self.Psi[0])\n",
    "            tmp = (self.Psi[0] * V_j0).crop(self.prec)\n",
    "            for i in range(1, len(self.Psi)):\n",
    "                V_ji = self.poisson(Phi[j] * self.Psi[i])\n",
    "                tmp += (self.Psi[i] * V_ji).crop(self.prec)\n",
    "            tmp *= 4.0*np.pi\n",
    "            Phi_out.append(tmp)\n",
    "        return np.array([phi.crop(self.prec) for phi in Phi_out])\n",
    "\n",
    "class KineticEnergyOperator():\n",
    "    def __init__(self, mra, prec):\n",
    "        self.prec = prec\n",
    "        self.mra = mra\n",
    "        self.derivative = vp.ABGVDerivative(mra, 0.5, 0.5)\n",
    "\n",
    "\n",
    "    def laplace_operator(self, phi):\n",
    "        return self.derivative(self.derivative(phi, 0), 0) + self.derivative(self.derivative(phi, 1), 1) + self.derivative(self.derivative(phi, 2), 2)\n",
    "\n",
    "    def __call__(self, Phi):\n",
    "        return np.array([(-0.5*self.laplace_operator(phi)).crop(self.prec) for phi in Phi])\n",
    "\n",
    "class CouloumbOperator():\n",
    "    \"\"\"\n",
    "    Vectorized Couloumb operator\n",
    "\n",
    "    J(phi) = \\sum_i P[psi_i * psi_i] * phi\n",
    "\n",
    "    Parameters\n",
    "    ----------\n",
    "    mra : The multiresolution analysis we work on\n",
    "    Psi : Orbital vector defining the operator\n",
    "    prec : Precision requirement\n",
    "\n",
    "    Attributes\n",
    "    ----------\n",
    "    poisson : Poisson operator\n",
    "    potential : Coulomb potential\n",
    "\n",
    "    \"\"\"\n",
    "    def __init__(self, mra, Psi, prec):\n",
    "        self.mra = mra\n",
    "        self.Psi = Psi\n",
    "        self.prec = prec\n",
    "        self.poisson = vp.PoissonOperator(mra=mra, prec=self.prec)\n",
    "        self.potential = None\n",
    "        self.setup()\n",
    "\n",
    "    def setup(self):\n",
    "        rho = self.Psi[0]**2\n",
    "        for i in range(1, len(self.Psi)):\n",
    "            rho += self.Psi[i]**2\n",
    "        rho.crop(self.prec)\n",
    "        self.potential = (4.0*np.pi)*self.poisson(rho).crop(self.prec)\n",
    "\n",
    "    def __call__(self, Phi):\n",
    "        \"\"\"Apply Couloumb operator onto an orbital vector Phi\"\"\"\n",
    "        return np.array([(self.potential*phi).crop(self.prec) for phi in Phi])\n",
    "\n",
    "\n",
    "\n",
    "class NuclearOperator():\n",
    "    \"\"\"\n",
    "    Vectorized Nuclear potential operator\n",
    "\n",
    "    Parameters\n",
    "    ----------\n",
    "    mra : The multiresolution analysis we work on\n",
    "    atoms : List of dicts containing charge and coordinates of the atoms\n",
    "    prec : Precision requirement\n",
    "\n",
    "    Attributes\n",
    "    ----------\n",
    "    potential : Nuclear potential\n",
    "\n",
    "    \"\"\"\n",
    "    def __init__(self, mra, atoms, prec):\n",
    "        self.mra = mra\n",
    "        self.prec= prec\n",
    "        self.atoms = atoms\n",
    "        self.potential = None\n",
    "        self.setup()\n",
    "\n",
    "    def setup(self):\n",
    "        # Project analytic function onto MRA basis\n",
    "        P_mra = vp.ScalingProjector(mra=self.mra, prec=self.prec)\n",
    "        f_nuc = NuclearFunction(self.atoms)\n",
    "        self.potential = P_mra(f_nuc)\n",
    "\n",
    "    def __call__(self, Phi):\n",
    "        \"\"\"Apply nuclear potential operator onto an orbital vector\"\"\"\n",
    "        return np.array([(self.potential*phi).crop(self.prec) for phi in Phi])\n",
    "\n",
    "class NuclearFunction():\n",
    "    \"\"\"\n",
    "    Vectorized Nuclear potential operator\n",
    "\n",
    "    Parameters\n",
    "    ----------\n",
    "\n",
    "    atoms : List of dicts containing charge and coordinates of the atoms\n",
    "\n",
    "    \"\"\"\n",
    "\n",
    "\n",
    "    def __init__(self, atoms):\n",
    "        self.atoms = atoms\n",
    "\n",
    "    def __call__(self, r):\n",
    "        \"Returns the nuclear potential value in R\"\n",
    "        tmp = 0.0\n",
    "        for atom in self.atoms:\n",
    "            Z = atom[\"Z\"]\n",
    "            R = atom[\"R\"]\n",
    "            R2 = (R[0]-r[0])**2 + (R[1]-r[1])**2 + (R[2]-r[2])**2\n",
    "            tmp += -Z / np.sqrt(R2)\n",
    "        return tmp\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [],
   "source": [
    "def scf_solver(mra, atoms, Phi_n, F_n, precision, threshold, max_iter=30):\n",
    "    \"\"\"Kinetric-free Hartree-Fock SCF solver\n",
    "\n",
    "    Parameters:\n",
    "    mra : The Multiresolution analysis to work on\n",
    "    atoms : List of dicts containing charge and coordinates of the atoms\n",
    "    Phi_n : Starting guess orbitals\n",
    "    F_n : Starting guess for Fock matrix\n",
    "    precision : Precision requirement\n",
    "    threshold : Usually set to the same as precision, set to -1 to limit iterations by max_iter\n",
    "    max_iter : Set maximum iterations\n",
    "\n",
    "    Returns:\n",
    "    Updates : Vector of orbital residual norms at each iteration\n",
    "    Energies : List of energy contributions at each iteration\n",
    "    Phi_n : Converged orbital vector\n",
    "\n",
    "    \"\"\"\n",
    "\n",
    "    # Setup nuclear potential\n",
    "    V_nuc = NuclearOperator(mra, atoms, precision)\n",
    "\n",
    "    # Loop parameters\n",
    "    iteration = 0                # Iteration counter\n",
    "    update = np.ones(len(Phi_n)) # Initialize error measure (norm of orbital updates)\n",
    "    updates = []                 # Will capture wavefunction updates for visusualization\n",
    "    energies = []                # Will capture energies for visualization\n",
    "\n",
    "    # SCF loop\n",
    "    while (max(update) > threshold):\n",
    "        if iteration > max_iter-1:\n",
    "            break\n",
    "\n",
    "        # Initialize operators for first iteration\n",
    "        J_n = CouloumbOperator(mra, Phi_n, precision)\n",
    "        K_n = ExchangeOperator(mra, Phi_n, precision)\n",
    "\n",
    "        # Initialize vector of Helmholtz operators based on Fock matrix diagonal\n",
    "        Lambda_n = np.diag(np.diag(F_n))\n",
    "        G = HelmholtzOperator(mra, np.diag(Lambda_n), precision)\n",
    "\n",
    "        # Apply potential operator to all orbitals\n",
    "        VPhi = V_nuc(Phi_n) + 2*J_n(Phi_n) - K_n(Phi_n)\n",
    "        KinOpr = KineticEnergyOperator(mra, precision)\n",
    "\n",
    "        # Apply Helmholtz operators to all orbitals\n",
    "        Phi_np1 = -2*G(VPhi + (Lambda_n - F_n) @ Phi_n)\n",
    "        dPhi_n = Phi_np1 - Phi_n\n",
    "        update = np.array([phi.norm() for phi in dPhi_n])\n",
    "\n",
    "        # Compute overlap matrices\n",
    "        S_tilde = calc_overlap(Phi_np1, Phi_np1)\n",
    "\n",
    "        # Löwdin orthonormalization S^{-1/2} = U * Sigma^{-1/2} * U^T\n",
    "        sigma, U = LA.eig(S_tilde)\n",
    "        Sm5 = U @ np.diag(sigma**(-0.5)) @ U.transpose()\n",
    "        Phi_bar = Sm5 @ Phi_np1\n",
    "\n",
    "        # Initialize n+1 operators\n",
    "        J_np1 = CouloumbOperator(mra, Phi_bar, precision)\n",
    "        K_np1 = ExchangeOperator(mra, Phi_bar, precision)\n",
    "\n",
    "        # Prepare for next iteration\n",
    "        Phi_n = Phi_bar\n",
    "        VPhikin = V_nuc(Phi_n) + 2*J_np1(Phi_n) - K_np1(Phi_n) + KinOpr(Phi_n)\n",
    "        F_n = calc_overlap(Phi_n, VPhikin)\n",
    "\n",
    "        # Compute energy contributions\n",
    "        energy = calc_energies(F_n, Phi_n, V_nuc, J_np1, K_np1)\n",
    "\n",
    "        # Collect output\n",
    "        updates.append(update)\n",
    "        energies.append(energy)\n",
    "        print(iteration, \" |  E_tot:\", energy[\"$E_{tot}$\"], \" |  dPhi:\", max(update))\n",
    "        iteration += 1\n",
    "\n",
    "\n",
    "    return np.array(updates), energies, Phi_n\n",
    "\n",
    "\n",
    "def calc_energies(F_mat, Phi, V, J, K):\n",
    "    \"\"\"\"Calcuate all energy contributions\"\"\"\n",
    "\n",
    "    V_mat = calc_overlap(Phi, V(Phi))\n",
    "    J_mat = calc_overlap(Phi, J(Phi))\n",
    "    K_mat = calc_overlap(Phi, K(Phi))\n",
    "\n",
    "    E_orb  = 2.0*F_mat.trace()\n",
    "    E_en   = 2.0*V_mat.trace()\n",
    "    E_coul = 2.0*J_mat.trace()\n",
    "    E_ex   = -K_mat.trace()\n",
    "    E_tot  = E_orb - E_coul - E_ex\n",
    "    E_kin  = E_tot - E_en - E_coul - E_ex\n",
    "\n",
    "    return {\n",
    "        \"$E_{orb}$\": E_orb,\n",
    "        \"$E_{en}$\": E_en,\n",
    "        \"$E_{coul}$\": E_coul,\n",
    "        \"$E_{ex}$\": E_ex,\n",
    "        \"$E_{kin}$\": E_kin,\n",
    "        \"$E_{tot}$\": E_tot\n",
    "    }\n",
    "\n",
    "\n",
    "def calc_overlap(Bra, Ket):\n",
    "    \"\"\"Calculate the overlap matrix between the orbitals <Bra| and |Ket>\n",
    "\n",
    "    Parameters:\n",
    "    Bra : bra vector <Phi|\n",
    "    Ket : ket vector |Phi>\n",
    "\n",
    "    Returns:\n",
    "    Overlap matrix\n",
    "\n",
    "    \"\"\"\n",
    "\n",
    "    S = np.empty((len(Bra), len(Ket)))\n",
    "    for i in range(len(Bra)):\n",
    "        for j in range(len(Ket)):\n",
    "            S[i, j] = vp.dot(Bra[i], Ket[j])\n",
    "    return S\n",
    "\n",
    "\n",
    "def starting_guess(mra, atom, n_orbs, prec):\n",
    "    \"\"\"Primitive starting guess, works for Be\"\"\"\n",
    "\n",
    "    # Define projector onto the MRA basis\n",
    "    P_mra = vp.ScalingProjector(mra=mra, prec=prec)\n",
    "\n",
    "    Phi = []\n",
    "    for i in range(1, n_orbs+1):\n",
    "        R0 = atom[\"R\"]\n",
    "        def f_gauss(r):\n",
    "            R2 = (r[0]-R0[0])**2 + (r[1]-R0[1])**2 + (r[2]-R0[2])**2\n",
    "            return np.exp(-R2/i)\n",
    "\n",
    "        phi = P_mra(f_gauss)\n",
    "        phi.normalize()\n",
    "        Phi.append(phi)\n",
    "    Phi = np.array(Phi)\n",
    "\n",
    "    # Löwdin orthonormalization S^{-1/2} = U * Sigma^{-1/2} * U^T\n",
    "    S = calc_overlap(Phi, Phi)\n",
    "    sigma, U = LA.eig(S)\n",
    "    Sm5 = U @ np.diag(sigma**(-0.5)) @ U.transpose()\n",
    "    return Sm5 @ Phi"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Define molecule\n",
    "mol = [\n",
    "    {\"Z\": 4.0, \"R\": [0.0, 0.0, 0.0]}\n",
    "]\n",
    "\n",
    "# Set target precision\n",
    "precision = 1.0e-3\n",
    "\n",
    "# Define MRA basis and computational domain\n",
    "k = 5                       # Polynomial order of MRA basis\n",
    "world = [-20, 20]           # Computational domain [-L,L]^3 (a.u.)\n",
    "MRA = vp.MultiResolutionAnalysis(order=k, box=world)\n",
    "\n",
    "# Setup guess for the orbitals and the Fock matrix\n",
    "Phi_n = starting_guess(mra=MRA, atom=mol[0], n_orbs=2, prec=precision)\n",
    "F_n = np.array([\n",
    "    [-1.0, 0.0],\n",
    "    [ 0.0,-0.5]\n",
    "]) "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "0  |  E_tot: -12.158806160497663  |  dPhi: 0.9554261103233852\n",
      "1  |  E_tot: -13.910545962062471  |  dPhi: 0.6030032083595198\n",
      "2  |  E_tot: -14.425154911621195  |  dPhi: 0.11248971009686566\n",
      "3  |  E_tot: -14.550559502300937  |  dPhi: 0.056367744794410625\n",
      "4  |  E_tot: -14.573091352701868  |  dPhi: 0.024464066721782796\n",
      "5  |  E_tot: -14.576656010668405  |  dPhi: 0.009607118621517596\n",
      "6  |  E_tot: -14.577184437878596  |  dPhi: 0.0036909439005734903\n",
      "7  |  E_tot: -14.577255599761454  |  dPhi: 0.0017015854922230098\n",
      "8  |  E_tot: -14.577262672068981  |  dPhi: 0.0009093846224262461\n"
     ]
    }
   ],
   "source": [
    "# Run SCF solver to optimize the orbitals\n",
    "orbital_updates, energies, Phi_n = scf_solver(MRA, mol, Phi_n, F_n, precision, precision, max_iter=30)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3.9.6 64-bit",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.6"
  },
  "orig_nbformat": 4,
  "vscode": {
   "interpreter": {
    "hash": "31f2aee4e71d21fbe5cf8b01ff0e069b9275f58929596ceb00d14d90e3e16cd6"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
